"""Patito, a data-modelling library built on top of polars and pydantic."""

from polars import Expr, Series, col

from patito import exceptions
from patito.exceptions import DataFrameValidationError
from patito.polars import DataFrame, LazyFrame
from patito.pydantic import Field, Model

_CACHING_AVAILABLE = False
field = col("_")
__all__ = [
    "DataFrame",
    "DataFrameValidationError",
    "Expr",
    "Field",
    "LazyFrame",
    "Model",
    "Series",
    "_CACHING_AVAILABLE",
    "col",
    "exceptions",
    "field",
]

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



import sys
import types
import typing
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

if typing.TYPE_CHECKING:
    Loc = Tuple[Union[int, str], ...]
    ReprArgs = Sequence[Tuple[Optional[str], Any]]
    RichReprResult = Iterable[
        Union[Any, Tuple[Any], Tuple[str, Any], Tuple[str, Any, Any]]
    ]

try:
    from typing import _TypingBase  # type: ignore[attr-defined]
except ImportError:
    from typing import _Final as _TypingBase  # type: ignore[attr-defined]

typing_base = _TypingBase

if sys.version_info < (3, 9):
    # python < 3.9 does not have GenericAlias (list[int], tuple[str, ...] and so on)
    TypingGenericAlias = ()
else:
    from typing import GenericAlias as TypingGenericAlias  # type: ignore

if sys.version_info < (3, 10):

    def origin_is_union(tp: Optional[Type[Any]]) -> bool:
        return tp is typing.Union

    WithArgsTypes = (TypingGenericAlias,)

else:

    def origin_is_union(tp: type[Any] | None) -> bool:
        return tp is typing.Union or tp is types.UnionType

    WithArgsTypes = typing._GenericAlias, types.GenericAlias, types.UnionType  # type: ignore[attr-defined]


class Representation:
    """Mixin to provide __str__, __repr__, and __pretty__ methods. See #884 for more details.

    __pretty__ is used by [devtools](https://python-devtools.helpmanual.io/) to provide human readable representations
    of objects.
    """

    __slots__: Tuple[str, ...] = tuple()

    def __repr_args__(self) -> "ReprArgs":
        """Returns the attributes to show in __str__, __repr__, and __pretty__ this is generally overridden.

        Can either return:
        * name - value pairs, e.g.: `[('foo_name', 'foo'), ('bar_name', ['b', 'a', 'r'])]`
        * or, just values, e.g.: `[(None, 'foo'), (None, ['b', 'a', 'r'])]`
        """
        attrs = ((s, getattr(self, s)) for s in self.__slots__)
        return [(a, v) for a, v in attrs if v is not None]

    def __repr_name__(self) -> str:
        """Name of the instance's class, used in __repr__."""
        return self.__class__.__name__

    def __repr_str__(self, join_str: str) -> str:
        return join_str.join(
            repr(v) if a is None else f"{a}={v!r}" for a, v in self.__repr_args__()
        )

    def __pretty__(
        self, fmt: Callable[[Any], Any], **kwargs: Any
    ) -> Generator[Any, None, None]:
        """Used by devtools (https://python-devtools.helpmanual.io/) to provide a human readable representations of objects."""
        yield self.__repr_name__() + "("
        yield 1
        for name, value in self.__repr_args__():
            if name is not None:
                yield name + "="
            yield fmt(value)
            yield ","
            yield 0
        yield -1
        yield ")"

    def __str__(self) -> str:
        return self.__repr_str__(" ")

    def __repr__(self) -> str:
        return f'{self.__repr_name__()}({self.__repr_str__(", ")})'

    def __rich_repr__(self) -> "RichReprResult":
        """Get fields for Rich library."""
        for name, field_repr in self.__repr_args__():
            if name is None:
                yield field_repr
            else:
                yield name, field_repr


def display_as_type(obj: Any) -> str:
    """Pretty representation of a type, should be as close as possible to the original type definition string.

    Takes some logic from `typing._type_repr`.
    """
    if isinstance(obj, types.FunctionType):
        return obj.__name__
    elif obj is ...:
        return "..."
    elif isinstance(obj, Representation):
        return repr(obj)

    if not isinstance(obj, (typing_base, WithArgsTypes, type)):
        obj = obj.__class__

    if origin_is_union(get_origin(obj)):
        args = ", ".join(map(display_as_type, get_args(obj)))
        return f"Union[{args}]"
    elif isinstance(obj, WithArgsTypes):
        if get_origin(obj) == Literal:
            args = ", ".join(map(repr, get_args(obj)))
        else:
            args = ", ".join(map(display_as_type, get_args(obj)))
        return f"{obj.__qualname__}[{args}]"
    elif isinstance(obj, type):
        return obj.__qualname__
    else:
        return repr(obj).replace("typing.", "").replace("typing_extensions.", "")



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Type, cast, get_args

from pydantic.fields import FieldInfo

from patito._pydantic.column_info import ColumnInfo
from patito._pydantic.dtypes import PYTHON_TO_PYDANTIC_TYPES

if TYPE_CHECKING:
    from patito.pydantic import ModelType


@cache
def schema_for_model(cls: Type[ModelType]) -> Dict[str, Dict[str, Any]]:
    """Return schema properties where definition references have been resolved.

    Returns:
        Field information as a dictionary where the keys are field names and the
            values are dictionaries containing metadata information about the field
            itself.

    Raises:
        TypeError: if a field is annotated with an enum where the values are of
            different types.

    """
    schema = cls.model_json_schema(by_alias=False, ref_template="{model}")
    fields = {}
    # first resolve definitions for nested models TODO checks for one-way references, if models are self-referencing this falls apart with recursion depth error
    for f in cls.model_fields.values():
        annotation = f.annotation
        cls._update_dfn(annotation, schema)
        for a in get_args(annotation):
            cls._update_dfn(a, schema)
    for field_name, field_info in schema["properties"].items():
        fields[field_name] = _append_field_info_to_props(
            field_info=field_info,
            field_name=field_name,
            required=field_name in schema.get("required", set()),
            model_schema=schema,
        )
    schema["properties"] = fields
    return schema


@cache
def column_infos_for_model(cls: Type[ModelType]) -> Mapping[str, ColumnInfo]:
    fields = cls.model_fields

    def get_column_info(field: FieldInfo) -> ColumnInfo:
        if field.json_schema_extra is None:
            return cast(ColumnInfo, cls.column_info_class())
        elif callable(field.json_schema_extra):
            raise NotImplementedError(
                "Callable json_schema_extra not supported by patito."
            )
        return cast(ColumnInfo, field.json_schema_extra["column_info"])

    return {k: get_column_info(v) for k, v in fields.items()}


def _append_field_info_to_props(
    field_info: Dict[str, Any],
    field_name: str,
    model_schema: Dict[str, Any],
    required: Optional[bool] = None,
) -> Dict[str, Any]:
    if "$ref" in field_info:  # TODO onto runtime append
        definition = model_schema["$defs"][field_info["$ref"]]
        if "enum" in definition and "type" not in definition:
            enum_types = set(type(value) for value in definition["enum"])
            if len(enum_types) > 1:
                raise TypeError(
                    "All enumerated values of enums used to annotate "
                    "Patito model fields must have the same type. "
                    "Encountered types: "
                    f"{sorted(map(lambda t: t.__name__, enum_types))}."
                )
            enum_type = enum_types.pop()
            definition["type"] = PYTHON_TO_PYDANTIC_TYPES[enum_type]
        field = definition
    else:
        field = field_info
    if "items" in field_info:
        field["items"] = _append_field_info_to_props(
            field_info=field_info["items"],
            field_name=field_name,
            model_schema=model_schema,
        )
    if required is not None:
        field["required"] = required
    if "const" in field_info and "type" not in field_info:
        field["type"] = PYTHON_TO_PYDANTIC_TYPES[type(field_info["const"])]
    return field



from __future__ import annotations

import json
from typing import (
    Any,
    Dict,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

import polars as pl
from polars.datatypes import DataType, DataTypeClass
from pydantic import BaseModel, field_serializer

from patito._pydantic.dtypes import parse_composite_dtype


class ColumnInfo(BaseModel, arbitrary_types_allowed=True):
    """patito-side model for storing column metadata.

    Args:
        constraints (Union[polars.Expression, List[polars.Expression]): A single
            constraint or list of constraints, expressed as a polars expression objects.
            All rows must satisfy the given constraint. You can refer to the given column
            with ``pt.field``, which will automatically be replaced with
            ``polars.col(<field_name>)`` before evaluation.
        derived_from (Union[str, polars.Expr]): used to mark fields that are meant to be derived from other fields. Users can specify a polars expression that will be called to derive the column value when `pt.DataFrame.derive` is called.
        dtype (polars.datatype.DataType): The given dataframe column must have the given
            polars dtype, for instance ``polars.UInt64`` or ``pl.Float32``.
        unique (bool): All row values must be unique.

    """

    dtype: Optional[Union[DataTypeClass, DataType]] = None
    constraints: Optional[Union[pl.Expr, Sequence[pl.Expr]]] = None
    derived_from: Optional[Union[str, pl.Expr]] = None
    unique: Optional[bool] = None

    def __repr__(self) -> str:
        """Print only Field attributes whose values are not default (mainly None)."""
        not_default_field = {
            field: getattr(self, field)
            for field in self.model_fields
            if getattr(self, field) is not self.model_fields[field].default
        }

        string = ""
        for field, value in not_default_field.items():
            string += f"{field}={value}, "
        if string:
            # remove trailing comma and space
            string = string[:-2]
        return f"ColumnInfo({string})"

    @field_serializer("constraints", "derived_from")
    def serialize_exprs(self, exprs: str | pl.Expr | Sequence[pl.Expr] | None) -> Any:
        if exprs is None:
            return None
        elif isinstance(exprs, str):
            return exprs
        elif isinstance(exprs, pl.Expr):
            return self._serialize_expr(exprs)
        elif isinstance(exprs, Sequence):
            return [self._serialize_expr(c) for c in exprs]
        else:
            raise ValueError(f"Invalid type for exprs: {type(exprs)}")

    def _serialize_expr(self, expr: pl.Expr) -> Dict:
        if isinstance(expr, pl.Expr):
            return json.loads(
                expr.meta.serialize(None)
            )  # can we access the dictionary directly?
        else:
            raise ValueError(f"Invalid type for expr: {type(expr)}")

    @field_serializer("dtype")
    def serialize_dtype(self, dtype: DataTypeClass | DataType | None) -> Any:
        """Serialize a polars dtype.

        References:
            [1] https://stackoverflow.com/questions/76572310/how-to-serialize-deserialize-polars-datatypes
        """
        if dtype is None:
            return None
        elif isinstance(dtype, DataTypeClass) or isinstance(dtype, DataType):
            return parse_composite_dtype(dtype)
        else:
            raise ValueError(f"Invalid type for dtype: {type(dtype)}")


CI = TypeVar("CI", bound=Type[ColumnInfo])



from __future__ import annotations

import sys
from enum import Enum
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
    cast,
    get_args,
    get_origin,
)

import polars as pl
from polars.datatypes import DataType, DataTypeClass, DataTypeGroup, convert
from polars.datatypes.constants import (
    DATETIME_DTYPES,
    DURATION_DTYPES,
    FLOAT_DTYPES,
    INTEGER_DTYPES,
)
from polars.polars import (
    dtype_str_repr,  # TODO: this is a rust function, can we implement our own string parser for Time/Duration/Datetime?
)

PYTHON_TO_PYDANTIC_TYPES = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    type(None): "null",
}

BOOLEAN_DTYPES = DataTypeGroup([pl.Boolean])
STRING_DTYPES = DataTypeGroup([pl.String])
DATE_DTYPES = DataTypeGroup([pl.Date])
TIME_DTYPES = DataTypeGroup([pl.Time])

PT_BASE_SUPPORTED_DTYPES = DataTypeGroup(
    INTEGER_DTYPES
    | FLOAT_DTYPES
    | BOOLEAN_DTYPES
    | STRING_DTYPES
    | DATE_DTYPES
    | DATETIME_DTYPES
    | DURATION_DTYPES
    | TIME_DTYPES
)

if sys.version_info >= (3, 10):  # pragma: no cover
    from types import UnionType  # pyright: ignore

    UNION_TYPES = (Union, UnionType)
else:
    UNION_TYPES = (Union,)  # pragma: no cover


class PydanticBaseType(Enum):
    STRING = "string"
    INTEGER = "integer"
    NUMBER = "number"
    BOOLEAN = "boolean"
    NULL = "null"
    OBJECT = "object"


class PydanticStringFormat(Enum):
    DATE = "date"
    DATE_TIME = "date-time"
    DURATION = "duration"
    TIME = "time"


def is_optional(type_annotation: type[Any] | Any | None) -> bool:
    """Return True if the given type annotation is an Optional annotation.

    Args:
        type_annotation: The type annotation to be checked.

    Returns:
        True if the outermost type is Optional.

    """
    return (get_origin(type_annotation) in UNION_TYPES) and (
        type(None) in get_args(type_annotation)
    )


def parse_composite_dtype(dtype: DataTypeClass | DataType) -> str:
    """For serialization, converts polars dtype to string representation."""
    if dtype in pl.NESTED_DTYPES:
        if dtype == pl.Struct or isinstance(dtype, pl.Struct):
            raise NotImplementedError("Structs not yet supported by patito")
        if not isinstance(dtype, pl.List) or isinstance(dtype, pl.Array):
            raise NotImplementedError(
                f"Unsupported nested dtype: {dtype} of type {type(dtype)}"
            )
        if dtype.inner is None:
            return convert.DataTypeMappings.DTYPE_TO_FFINAME[dtype.base_type()]
        return f"{convert.DataTypeMappings.DTYPE_TO_FFINAME[dtype.base_type()]}[{parse_composite_dtype(dtype.inner)}]"
    elif dtype in pl.TEMPORAL_DTYPES:
        return cast(str, dtype_str_repr(dtype))
    else:
        return convert.DataTypeMappings.DTYPE_TO_FFINAME[dtype]


def dtype_from_string(v: str) -> Optional[Union[DataTypeClass, DataType]]:
    """For deserialization."""
    # TODO test all dtypes
    return convert.dtype_short_repr_to_dtype(v)


def _pyd_type_to_valid_dtypes(
    pyd_type: PydanticBaseType, string_format: Optional[str], enum: List[str] | None
) -> DataTypeGroup:
    if enum is not None:
        _validate_enum_values(pyd_type, enum)
        return DataTypeGroup([pl.Enum(enum), pl.String], match_base_type=False)
    if pyd_type.value == "integer":
        return DataTypeGroup(INTEGER_DTYPES | FLOAT_DTYPES)
    elif pyd_type.value == "number":
        return (
            FLOAT_DTYPES
            if isinstance(FLOAT_DTYPES, DataTypeGroup)
            else DataTypeGroup(FLOAT_DTYPES)
        )
    elif pyd_type.value == "boolean":
        return BOOLEAN_DTYPES
    elif pyd_type.value == "string":
        _string_format = (
            PydanticStringFormat(string_format) if string_format is not None else None
        )
        return _pyd_string_format_to_valid_dtypes(_string_format)
    elif pyd_type.value == "null":
        return DataTypeGroup([pl.Null])
    else:
        return DataTypeGroup([])


def _pyd_type_to_default_dtype(
    pyd_type: PydanticBaseType, string_format: Optional[str], enum: List[str] | None
) -> DataTypeClass | DataType:
    if enum is not None:
        _validate_enum_values(pyd_type, enum)
        return pl.Enum(enum)
    elif pyd_type.value == "integer":
        return pl.Int64()
    elif pyd_type.value == "number":
        return pl.Float64()
    elif pyd_type.value == "boolean":
        return pl.Boolean()
    elif pyd_type.value == "string":
        _string_format = (
            PydanticStringFormat(string_format) if string_format is not None else None
        )
        return _pyd_string_format_to_default_dtype(_string_format)
    elif pyd_type.value == "null":
        return pl.Null()
    elif pyd_type.value == "object":
        raise ValueError("pydantic object types not currently supported by patito")
    else:
        raise NotImplementedError


def _pyd_string_format_to_valid_dtypes(
    string_format: PydanticStringFormat | None,
) -> DataTypeGroup:
    if string_format is None:
        return STRING_DTYPES
    elif string_format.value == "date":
        return DATE_DTYPES
    elif string_format.value == "date-time":
        return (
            DATETIME_DTYPES
            if isinstance(DATE_DTYPES, DataTypeGroup)
            else DataTypeGroup(DATE_DTYPES)
        )
    elif string_format.value == "duration":
        return (
            DURATION_DTYPES
            if isinstance(DURATION_DTYPES, DataTypeGroup)
            else DataTypeGroup(DURATION_DTYPES)
        )
    elif string_format.value == "time":
        return TIME_DTYPES
    else:
        raise NotImplementedError


def _pyd_string_format_to_default_dtype(
    string_format: PydanticStringFormat | None,
) -> DataTypeClass | DataType:
    if string_format is None:
        return pl.String()
    elif string_format.value == "date":
        return pl.Date()
    elif string_format.value == "date-time":
        return pl.Datetime()
    elif string_format.value == "duration":
        return pl.Duration()
    elif string_format.value == "time":
        return pl.Time()
    else:
        raise NotImplementedError


def _without_optional(schema: Dict) -> Dict:
    if "anyOf" in schema:
        for sub_props in schema["anyOf"]:
            if "type" in sub_props and sub_props["type"] == "null":
                schema["anyOf"].remove(sub_props)
    return schema


def _validate_enum_values(pyd_type: PydanticBaseType, enum: Sequence[Any]) -> None:
    enum_types = set(type(value) for value in enum)
    if len(enum_types) > 1:
        raise TypeError(
            f"All enumerated values of enums used to annotate Patito model fields must have the same type. Encountered types: {sorted(map(lambda t: t.__name__, enum_types))}."
        )
    if pyd_type.value != "string":
        raise TypeError(
            f"Enums used to annotate Patito model fields must be strings. Encountered type: {enum_types.pop().__name__}."
        )



from patito._pydantic.dtypes.dtypes import (
    DtypeResolver,
    default_dtypes_for_model,
    valid_dtypes_for_model,
    validate_annotation,
    validate_polars_dtype,
)
from patito._pydantic.dtypes.utils import (
    PYTHON_TO_PYDANTIC_TYPES,
    dtype_from_string,
    is_optional,
    parse_composite_dtype,
)

__all__ = [
    "DtypeResolver",
    "validate_annotation",
    "validate_polars_dtype",
    "parse_composite_dtype",
    "dtype_from_string",
    "valid_dtypes_for_model",
    "default_dtypes_for_model",
    "PYTHON_TO_PYDANTIC_TYPES",
    "is_optional",
]



"""Exceptions used by patito."""

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    Union,
)

from patito._pydantic.repr import Representation

if TYPE_CHECKING:
    from pydantic import BaseModel

    Loc = Tuple[Union[int, str], ...]

    class _ErrorDictRequired(TypedDict):
        loc: Loc
        msg: str
        type: str

    class ErrorDict(_ErrorDictRequired, total=False):
        ctx: Dict[str, Any]

    from patito._pydantic.repr import ReprArgs


__all__ = "ErrorWrapper", "DataFrameValidationError"


class ErrorWrapper(Representation):
    """Error handler for nicely accumulating errors."""

    __slots__ = "exc", "_loc"

    def __init__(self, exc: Exception, loc: Union[str, "Loc"]) -> None:
        """Wrap an error in an ErrorWrapper."""
        self.exc = exc
        self._loc = loc

    def loc_tuple(self) -> "Loc":
        """Represent error as tuple."""
        if isinstance(self._loc, tuple):
            return self._loc
        else:
            return (self._loc,)

    def __repr_args__(self) -> "ReprArgs":
        """Pydantic repr."""
        return [("exc", self.exc), ("loc", self.loc_tuple())]


# ErrorList is something like Union[List[Union[List[ErrorWrapper], ErrorWrapper]], ErrorWrapper]
# but recursive, therefore just use:
ErrorList = Union[Sequence[Any], ErrorWrapper]


class DataFrameValidationError(Representation, ValueError):
    """Parent error for DataFrame validation errors."""

    __slots__ = "raw_errors", "model", "_error_cache"

    def __init__(self, errors: Sequence[ErrorList], model: Type["BaseModel"]) -> None:
        """Create a dataframe validation error."""
        self.raw_errors = errors
        self.model = model
        self._error_cache: Optional[List["ErrorDict"]] = None

    def errors(self) -> List["ErrorDict"]:
        """Get list of errors."""
        if self._error_cache is None:
            self._error_cache = list(flatten_errors(self.raw_errors))
        return self._error_cache

    def __str__(self) -> str:
        """String reprentation of error."""
        errors = self.errors()
        no_errors = len(errors)
        return (
            f'{no_errors} validation error{"" if no_errors == 1 else "s"} for {self.model.__name__}\n'
            f"{display_errors(errors)}"
        )

    def __repr_args__(self) -> "ReprArgs":
        """Pydantic repr."""
        return [("model", self.model.__name__), ("errors", self.errors())]


def display_errors(errors: List["ErrorDict"]) -> str:
    return "\n".join(
        f'{_display_error_loc(e)}\n  {e["msg"]} ({_display_error_type_and_ctx(e)})'
        for e in errors
    )


def _display_error_loc(error: "ErrorDict") -> str:
    return " -> ".join(str(e) for e in error["loc"])


def _display_error_type_and_ctx(error: "ErrorDict") -> str:
    t = "type=" + error["type"]
    ctx = error.get("ctx")
    if ctx:
        return t + "".join(f"; {k}={v}" for k, v in ctx.items())
    else:
        return t


def flatten_errors(
    errors: Sequence[Any], loc: Optional["Loc"] = None
) -> Generator["ErrorDict", None, None]:
    for error in errors:
        if isinstance(error, ErrorWrapper):
            if loc:
                error_loc = loc + error.loc_tuple()
            else:
                error_loc = error.loc_tuple()

            if isinstance(error.exc, DataFrameValidationError):
                yield from flatten_errors(error.exc.raw_errors, error_loc)
            else:
                yield error_dict(error.exc, error_loc)
        elif isinstance(error, list):
            yield from flatten_errors(error, loc=loc)
        else:
            raise RuntimeError(f"Unknown error object: {error}")


def error_dict(exc: Exception, loc: "Loc") -> "ErrorDict":
    type_ = get_exc_type(exc.__class__)
    msg_template = getattr(exc, "msg_template", None)
    ctx = exc.__dict__
    if msg_template:
        msg = msg_template.format(**ctx)
    else:
        msg = str(exc)

    d: "ErrorDict" = {"loc": loc, "msg": msg, "type": type_}

    if ctx:
        d["ctx"] = ctx

    return d


_EXC_TYPE_CACHE: Dict[Type[Exception], str] = {}


def get_exc_type(cls: Type[Exception]) -> str:
    # slightly more efficient than using lru_cache since we don't need to worry about the cache filling up
    try:
        return _EXC_TYPE_CACHE[cls]
    except KeyError:
        r = _get_exc_type(cls)
        _EXC_TYPE_CACHE[cls] = r
        return r


def _get_exc_type(cls: Type[Exception]) -> str:
    if issubclass(cls, AssertionError):
        return "assertion_error"

    base_name = "type_error" if issubclass(cls, TypeError) else "value_error"
    if cls in (TypeError, ValueError):
        # just TypeError or ValueError, no extra code
        return base_name

    # if it's not a TypeError or ValueError, we just take the lowercase of the exception name
    # no chaining or snake case logic, use "code" for more complex error types.
    code = getattr(cls, "code", None) or cls.__name__.replace("Error", "").lower()
    return base_name + "." + code


class WrongColumnsError(TypeError):
    """Validation exception for column name mismatches."""


class MissingColumnsError(WrongColumnsError):
    """Exception for when a dataframe is missing one or more columns."""


class SuperfluousColumnsError(WrongColumnsError):
    """Exception for when a dataframe has one ore more non-specified columns."""


class MissingValuesError(ValueError):
    """Exception for when a dataframe has non-nullable columns with nulls."""


class ColumnDTypeError(TypeError):
    """Exception for when a dataframe has one or more columns with wrong dtypes."""


class RowValueError(ValueError):
    """Exception for when a dataframe has a row with a impermissible value."""


class RowDoesNotExist(RuntimeError):
    """Exception for when a single row was expected, but none were returned."""


class MultipleRowsReturned(RuntimeError):
    """Exception for when a single row was expected, but several were returned."""



from __future__ import annotations

from functools import cache, reduce
from operator import and_
from typing import TYPE_CHECKING, Any, Dict, FrozenSet, Mapping, Optional, Type

import polars as pl
from polars.datatypes import DataType, DataTypeClass, DataTypeGroup
from pydantic import TypeAdapter

from patito._pydantic.dtypes.utils import (
    PT_BASE_SUPPORTED_DTYPES,
    PydanticBaseType,
    _pyd_type_to_default_dtype,
    _pyd_type_to_valid_dtypes,
    _without_optional,
    dtype_from_string,
)
from patito._pydantic.repr import display_as_type

if TYPE_CHECKING:
    from patito.pydantic import ModelType


@cache
def valid_dtypes_for_model(
    cls: Type[ModelType],
) -> Mapping[str, FrozenSet[DataTypeClass]]:
    return {
        column: (
            DtypeResolver(cls.model_fields[column].annotation).valid_polars_dtypes()
            if cls.column_infos[column].dtype is None
            else DataTypeGroup([cls.dtypes[column]], match_base_type=False)
        )
        for column in cls.columns
    }


@cache
def default_dtypes_for_model(
    cls: Type[ModelType],
) -> dict[str, DataType]:
    default_dtypes: dict[str, DataType] = {}
    for column in cls.columns:
        dtype = (
            cls.column_infos[column].dtype
            or DtypeResolver(cls.model_fields[column].annotation).default_polars_dtype()
        )
        if dtype is None:
            raise ValueError(f"Unable to find a default dtype for column `{column}`")

        default_dtypes[column] = dtype if isinstance(dtype, DataType) else dtype()
    return default_dtypes


def validate_polars_dtype(
    annotation: type[Any] | None,
    dtype: DataType | DataTypeClass | None,
    column: Optional[str] = None,
) -> None:
    """Check that the polars dtype is valid for the given annotation. Raises ValueError if not.

    Args:
        annotation (type[Any] | None): python type annotation
        dtype (DataType | DataTypeClass | None): polars dtype
        column (Optional[str], optional): column name. Defaults to None.

    """
    if (
        dtype is None or annotation is None
    ):  # no potential conflict between type annotation and chosen polars type
        return
    valid_dtypes = DtypeResolver(annotation).valid_polars_dtypes()
    if dtype not in valid_dtypes:
        if column:
            column_msg = f" for column `{column}`"
        else:
            column_msg = ""
        raise ValueError(
            f"Invalid dtype {dtype}{column_msg}. Allowable polars dtypes for {display_as_type(annotation)} are: {', '.join([str(x) for x in valid_dtypes])}."
        )
    return


def validate_annotation(
    annotation: type[Any] | Any | None, column: Optional[str] = None
) -> None:
    """Check that the provided annotation has polars/patito support (we can resolve it to a default dtype). Raises ValueError if not.

    Args:
        annotation (type[Any] | None): python type annotation
        column (Optional[str], optional): column name. Defaults to None.

    """
    default_dtype = DtypeResolver(annotation).default_polars_dtype()
    if default_dtype is None:
        valid_polars_dtypes = DtypeResolver(annotation).valid_polars_dtypes()
        if column:
            column_msg = f" for column `{column}`"
        else:
            column_msg = ""
        if len(valid_polars_dtypes) == 0:
            raise ValueError(
                f"Annotation {display_as_type(annotation)}{column_msg} is not compatible with any polars dtypes."
            )
        else:
            raise ValueError(
                f"Unable to determine default dtype for annotation {display_as_type(annotation)}{column_msg}. Please provide a valid default polars dtype via the `dtype` argument to `Field`. Valid dtypes are: {', '.join([str(x) for x in valid_polars_dtypes])}."
            )
    return


class DtypeResolver:
    def __init__(self, annotation: Any | None):
        self.annotation = annotation
        self.schema = TypeAdapter(annotation).json_schema()
        self.defs = self.schema.get("$defs", {})

    def valid_polars_dtypes(self) -> DataTypeGroup:
        if self.annotation == Any:
            return PT_BASE_SUPPORTED_DTYPES
        return self._valid_polars_dtypes_for_schema(self.schema)

    def default_polars_dtype(self) -> DataType | None:
        if self.annotation == Any:
            return pl.String()
        return self._default_polars_dtype_for_schema(self.schema)

    def _valid_polars_dtypes_for_schema(
        self,
        schema: Dict,
    ) -> DataTypeGroup:
        valid_type_sets = []
        if "anyOf" in schema:
            schema = _without_optional(schema)
            for sub_props in schema["anyOf"]:
                valid_type_sets.append(
                    self._pydantic_subschema_to_valid_polars_types(sub_props)
                )
        else:
            valid_type_sets.append(
                self._pydantic_subschema_to_valid_polars_types(schema)
            )
        return reduce(and_, valid_type_sets) if valid_type_sets else DataTypeGroup([])

    def _pydantic_subschema_to_valid_polars_types(
        self,
        props: Dict,
    ) -> DataTypeGroup:
        if "type" not in props:
            if "enum" in props:
                raise TypeError("Mixed type enums not supported by patito.")
            elif "const" in props:
                return DtypeResolver(type(props["const"])).valid_polars_dtypes()
            elif "$ref" in props:
                return self._pydantic_subschema_to_valid_polars_types(
                    self.defs[props["$ref"].split("/")[-1]]
                )
            return DataTypeGroup([])
        pyd_type = props.get("type")
        if pyd_type == "array":
            if "items" not in props:
                return DataTypeGroup([])
            array_props = props["items"]
            item_dtypes = self._valid_polars_dtypes_for_schema(array_props)
            # TODO support pl.Array?
            return DataTypeGroup(
                [pl.List(dtype) for dtype in item_dtypes], match_base_type=False
            )
        elif pyd_type == "object":
            if "properties" not in props:
                return DataTypeGroup([])
            object_props = props["properties"]
            return DataTypeGroup(
                [
                    pl.Struct(
                        [
                            pl.Field(
                                name, self._default_polars_dtype_for_schema(sub_props)
                            )
                            for name, sub_props in object_props.items()
                        ]
                    )
                ],
                match_base_type=False,
            )  # for structs, return only the default dtype set to avoid combinatoric issues
        return _pyd_type_to_valid_dtypes(
            PydanticBaseType(pyd_type), props.get("format"), props.get("enum")
        )

    def _default_polars_dtype_for_schema(self, schema: Dict) -> DataType | None:
        if "anyOf" in schema:
            if len(schema["anyOf"]) == 2:  # look for optionals first
                schema = _without_optional(schema)
            if len(schema["anyOf"]) == 1:
                if "column_info" in schema:
                    schema["anyOf"][0]["column_info"] = schema[
                        "column_info"
                    ]  # push column info through optional
                schema = schema["anyOf"][0]
            else:
                return None
        return self._pydantic_subschema_to_default_dtype(schema)

    def _pydantic_subschema_to_default_dtype(
        self,
        props: Dict,
    ) -> DataType | None:
        if "column_info" in props:  # user has specified in patito model
            if props["column_info"]["dtype"] is not None:
                dtype = dtype_from_string(props["column_info"]["dtype"])
                dtype = dtype() if isinstance(dtype, DataTypeClass) else dtype
                return dtype
        if "type" not in props:
            if "enum" in props:
                raise TypeError("Mixed type enums not supported by patito.")
            elif "const" in props:
                return DtypeResolver(type(props["const"])).default_polars_dtype()
            elif "$ref" in props:
                return self._pydantic_subschema_to_default_dtype(
                    self.defs[props["$ref"].split("/")[-1]]
                )
            return None
        pyd_type = props.get("type")
        if pyd_type == "array":
            if "items" not in props:
                raise NotImplementedError(
                    "Unexpected error processing pydantic schema. Please file an issue."
                )
            array_props = props["items"]
            inner_default_type = self._default_polars_dtype_for_schema(array_props)
            if inner_default_type is None:
                return None
            return pl.List(inner_default_type)
        elif pyd_type == "object":
            if "properties" not in props:
                raise NotImplementedError(
                    "dictionaries not currently supported by patito"
                )
            object_props = props["properties"]
            return pl.Struct(
                [
                    pl.Field(name, self._default_polars_dtype_for_schema(sub_props))
                    for name, sub_props in object_props.items()
                ]
            )
        return _pyd_type_to_default_dtype(
            PydanticBaseType(pyd_type), props.get("format"), props.get("enum")
        )



"""Module for validating datastructures with respect to model specifications."""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Sequence,
    Type,
    Union,
    _UnionGenericAlias,
    cast,
)

import polars as pl
from pydantic.aliases import AliasGenerator
from typing_extensions import get_args

from patito._pydantic.dtypes import is_optional
from patito.exceptions import (
    ColumnDTypeError,
    DataFrameValidationError,
    ErrorWrapper,
    MissingColumnsError,
    MissingValuesError,
    RowValueError,
    SuperfluousColumnsError,
)

try:
    import pandas as pd  # type: ignore

    _PANDAS_AVAILABLE = True
except ImportError:
    _PANDAS_AVAILABLE = False

if TYPE_CHECKING:
    from patito import Model


VALID_POLARS_TYPES = {
    "enum": {pl.Categorical},
    "boolean": {pl.Boolean},
    "string": {pl.String, pl.Datetime, pl.Date},
    "number": {pl.Float32, pl.Float64},
    "integer": {
        pl.Int8,
        pl.Int16,
        pl.Int32,
        pl.Int64,
        pl.UInt8,
        pl.UInt16,
        pl.UInt32,
        pl.UInt64,
    },
}


def _dewrap_optional(type_annotation: Type[Any] | Any) -> Type:
    """Return the inner, wrapped type of an Optional.

    Is a no-op for non-Optional types.

    Args:
        type_annotation: The type annotation to be dewrapped.

    Returns:
        The input type, but with the outermost Optional removed.

    """
    return (
        next(  # pragma: no cover
            valid_type
            for valid_type in get_args(type_annotation)
            if valid_type is not type(None)  # noqa: E721
        )
        if is_optional(type_annotation)
        else type_annotation
    )


def _transform_df(dataframe: pl.DataFrame, schema: type[Model]) -> pl.DataFrame:
    """Transform any properties of the dataframe according to the model.

    Currently only supports using AliasGenerator to transform column names to match a model.

    Args:
        dataframe: Polars DataFrame to be validated.
        schema: Patito model which specifies how the dataframe should be structured.

    """
    # Check if an alias generator is present in model_config
    if alias_gen := schema.model_config.get("alias_generator"):
        if isinstance(alias_gen, AliasGenerator):
            alias_func = alias_gen.validation_alias or alias_gen.alias
            assert (
                alias_func is not None
            ), "An AliasGenerator must contain a transforming function"
        else:  # alias_gen is a function
            alias_func = alias_gen

        new_cols: list[str] = [
            alias_func(field_name) for field_name in dataframe.columns
        ]  # type: ignore
        dataframe.columns = new_cols
    return dataframe


def _find_errors(  # noqa: C901
    dataframe: pl.DataFrame,
    schema: Type[Model],
    columns: Optional[Sequence[str]] = None,
    allow_missing_columns: bool = False,
    allow_superfluous_columns: bool = False,
) -> list[ErrorWrapper]:
    """Validate the given dataframe.

    Args:
        dataframe: Polars DataFrame to be validated.
        schema: Patito model which specifies how the dataframe should be structured.
        columns: If specified, only validate the given columns. Missing columns will
            check if any specified columns are missing from the inputted dataframe,
            and superfluous columns will check if any columns not specified in the
            schema are present in the columns list.
        allow_missing_columns: If True, missing columns will not be considered an error.
        allow_superfluous_columns: If True, additional columns will not be considered an error.

    Returns:
        A list of patito.exception.ErrorWrapper instances. The specific validation
        error can be retrieved from the "exc" attribute on each error wrapper instance.

        MissingColumnsError: If there are any missing columns.
        SuperfluousColumnsError: If there are additional, non-specified columns.
        MissingValuesError: If there are nulls in a non-optional column.
        ColumnDTypeError: If any column has the wrong dtype.
        NotImplementedError: If validation has not been implement for the given
            type.

    """
    errors: list[ErrorWrapper] = []
    schema_subset = columns or schema.columns
    column_subset = columns or dataframe.columns
    if not allow_missing_columns:
        # Check if any columns are missing
        for missing_column in set(schema_subset) - set(dataframe.columns):
            errors.append(
                ErrorWrapper(
                    MissingColumnsError("Missing column"),
                    loc=missing_column,
                )
            )

    if not allow_superfluous_columns:
        # Check if any additional columns are included
        for superfluous_column in set(column_subset) - set(schema.columns):
            errors.append(
                ErrorWrapper(
                    SuperfluousColumnsError("Superfluous column"),
                    loc=superfluous_column,
                )
            )

    # Check if any non-optional columns have null values
    for column in schema.non_nullable_columns.intersection(column_subset):
        num_missing_values = dataframe.get_column(name=column).null_count()
        if num_missing_values:
            errors.append(
                ErrorWrapper(
                    MissingValuesError(
                        f"{num_missing_values} missing "
                        f"{'value' if num_missing_values == 1 else 'values'}"
                    ),
                    loc=column,
                )
            )

    for column, dtype in schema.dtypes.items():
        if column not in column_subset:
            continue
        if not isinstance(dtype, pl.List):
            continue

        annotation = schema.model_fields[column].annotation  # type: ignore[unreachable]

        # Retrieve the annotation of the list itself,
        # dewrapping any potential Optional[...]
        list_type = _dewrap_optional(annotation)

        # Check if the list items themselves should be considered nullable
        item_type = get_args(list_type)[0]
        if is_optional(item_type):
            continue

        num_missing_values = (
            dataframe.lazy()
            .select(column)
            # Remove those rows that do not contain lists at all
            .filter(pl.col(column).is_not_null())
            # Convert lists of N items to N individual rows
            .explode(column)
            # Calculate how many nulls are present in lists
            .filter(pl.col(column).is_null())
            .collect()
            .height
        )
        if num_missing_values != 0:
            errors.append(
                ErrorWrapper(
                    MissingValuesError(
                        f"{num_missing_values} missing "
                        f"{'value' if num_missing_values == 1 else 'values'} "
                        f"in lists"
                    ),
                    loc=column,
                )
            )

    # Check if any column has a wrong dtype
    valid_dtypes = schema.valid_dtypes
    dataframe_datatypes = dict(zip(dataframe.columns, dataframe.dtypes))
    for column_name, column_properties in schema._schema_properties().items():
        column_info = schema.column_infos[column_name]
        if column_name not in dataframe.columns or column_name not in column_subset:
            continue

        polars_type = dataframe_datatypes[column_name]
        if polars_type not in valid_dtypes[column_name]:
            errors.append(
                ErrorWrapper(
                    ColumnDTypeError(
                        f"Polars dtype {polars_type} does not match model field type."
                    ),
                    loc=column_name,
                )
            )

        # Test for when only specific values are accepted
        e = _find_enum_errors(
            df=dataframe,
            column_name=column_name,
            props=column_properties,
            schema=schema,
        )
        if e is not None:
            errors.append(e)

        if column_info.unique:
            # Coalescing to 0 in the case of dataframe of height 0
            num_duplicated = dataframe[column_name].is_duplicated().sum() or 0
            if num_duplicated > 0:
                errors.append(
                    ErrorWrapper(
                        RowValueError(f"{num_duplicated} rows with duplicated values."),
                        loc=column_name,
                    )
                )

        # Intercept struct columns, and process errors separately
        if schema.dtypes[column_name] == pl.Struct:
            nested_schema = schema.model_fields[column_name].annotation

            # Additional unpack required if structs column is optional
            if type(nested_schema) == _UnionGenericAlias:
                nested_schema = nested_schema.__args__[0]

                # We need to filter out any null rows as the submodel won't know
                # that all of a row's columns may be null
                dataframe = dataframe.filter(pl.col(column_name).is_not_null())
                if dataframe.is_empty():
                    continue

            struct_errors = _find_errors(
                dataframe=dataframe.select(column_name).unnest(column_name),
                schema=nested_schema,
            )

            # Format nested errors
            for error in struct_errors:
                error._loc = f"{column_name}.{error._loc}"

            errors.extend(struct_errors)

            # No need to do any more checks
            continue

        # Intercept list of structs columns, and process errors separately
        elif schema.dtypes[column_name] == pl.List(pl.Struct):
            nested_schema = schema.model_fields[column_name].annotation.__args__[0]

            # Additional unpack required if structs column is optional
            if type(nested_schema) == _UnionGenericAlias:
                nested_schema = nested_schema.__args__[0]

                # We need to filter out any null rows as the submodel won't know
                # that all of a row's columns may be null
                dataframe = dataframe.filter(pl.col(column_name).is_not_null())
                if dataframe.is_empty():
                    continue

            list_struct_errors = _find_errors(
                dataframe=dataframe.select(column_name)
                .explode(column_name)
                .unnest(column_name),
                schema=nested_schema,
            )

            # Format nested errors
            for error in list_struct_errors:
                error._loc = f"{column_name}.{error._loc}"

            errors.extend(list_struct_errors)

            # No need to do any more checks
            continue

        # Check for bounded value fields
        col = pl.col(column_name)
        filters = {
            "maximum": lambda v, col=col: col <= v,
            "exclusiveMaximum": lambda v, col=col: col < v,
            "minimum": lambda v, col=col: col >= v,
            "exclusiveMinimum": lambda v, col=col: col > v,
            "multipleOf": lambda v, col=col: (col == 0) | ((col % v) == 0),
            "const": lambda v, col=col: col == v,
            "pattern": lambda v, col=col: col.str.contains(v),
            "minLength": lambda v, col=col: col.str.len_chars() >= v,
            "maxLength": lambda v, col=col: col.str.len_chars() <= v,
        }
        if "anyOf" in column_properties:
            checks = [
                check(x[key])
                for key, check in filters.items()
                for x in column_properties["anyOf"]
                if key in x
            ]
        else:
            checks = []
        checks += [
            check(column_properties[key])
            for key, check in filters.items()
            if key in column_properties
        ]
        if checks:
            n_invalid_rows = 0
            for check in checks:
                lazy_df = dataframe.lazy()
                lazy_df = lazy_df.filter(
                    ~check
                )  # get failing rows (nulls will evaluate to null on boolean check, we only want failures (false)))
                invalid_rows = lazy_df.collect()
                n_invalid_rows += invalid_rows.height
            if n_invalid_rows > 0:
                errors.append(
                    ErrorWrapper(
                        RowValueError(
                            f"{n_invalid_rows} row{'' if n_invalid_rows == 1 else 's'} "
                            "with out of bound values."
                        ),
                        loc=column_name,
                    )
                )

        if column_info.constraints is not None:
            custom_constraints = column_info.constraints
            if isinstance(custom_constraints, pl.Expr):
                custom_constraints = [custom_constraints]
            constraints = pl.any_horizontal(
                [constraint.not_() for constraint in custom_constraints]
            )
            if "_" in constraints.meta.root_names():
                # An underscore is an alias for the current field
                illegal_rows = dataframe.with_columns(
                    pl.col(column_name).alias("_")
                ).filter(constraints)
            else:
                illegal_rows = dataframe.filter(constraints)
            if illegal_rows.height > 0:
                errors.append(
                    ErrorWrapper(
                        RowValueError(
                            f"{illegal_rows.height} "
                            f"row{'' if illegal_rows.height == 1 else 's'} "
                            "does not match custom constraints."
                        ),
                        loc=column_name,
                    )
                )

    return errors


def _find_enum_errors(
    df: pl.DataFrame, column_name: str, props: dict[str, Any], schema: Type[Model]
) -> ErrorWrapper | None:
    if "enum" not in props:
        if "items" in props and "enum" in props["items"]:
            return _find_enum_errors(df, column_name, props["items"], schema)
        return None
    permissible_values = set(props["enum"])
    if column_name in schema.nullable_columns:
        permissible_values.add(None)
    if isinstance(df[column_name].dtype, pl.List):
        actual_values = set(df[column_name].explode().unique())
    else:
        actual_values = set(df[column_name].unique())
    impermissible_values = actual_values - permissible_values
    if impermissible_values:
        return ErrorWrapper(
            RowValueError(f"Rows with invalid values: {impermissible_values}."),
            loc=column_name,
        )
    return None


def validate(
    dataframe: Union["pd.DataFrame", pl.DataFrame],
    schema: Type[Model],
    columns: Optional[Sequence[str]] = None,
    allow_missing_columns: bool = False,
    allow_superfluous_columns: bool = False,
) -> None:
    """Validate the given dataframe.

    Args:
        dataframe: Polars DataFrame to be validated.
        schema: Patito model which specifies how the dataframe should be structured.
        columns: Optional list of columns to validate. If not provided, all columns
            of the dataframe will be validated.
        allow_missing_columns: If True, missing columns will not be considered an error.
        allow_superfluous_columns: If True, additional columns will not be considered an error.

    Raises:
        DataFrameValidationError: If the given dataframe does not match the given schema.

    """
    if _PANDAS_AVAILABLE and isinstance(dataframe, pd.DataFrame):
        polars_dataframe = pl.from_pandas(dataframe)
    else:
        polars_dataframe = cast(pl.DataFrame, dataframe)

    polars_dataframe = _transform_df(polars_dataframe, schema)
    errors = _find_errors(
        dataframe=polars_dataframe,
        schema=schema,
        columns=columns,
        allow_missing_columns=allow_missing_columns,
        allow_superfluous_columns=allow_superfluous_columns,
    )
    if errors:
        raise DataFrameValidationError(errors=errors, model=schema)
